/*
 * Jay Krell
 * jay.krell@cornell.edu
 * 1996
 */

#ifdef _WIN32

// rough history
/*
* Still exports extern "C", but uses some C++ features internally.
* Correctly handles very long UNC paths. Not tested, since these
* long paths are difficult to create. (Tested with unc, just not
* unc > 260)
*
* Completely reformatted braces. Sorry.
*
* Put in some tentative translations for errno. This probably will
* *not* be fleshed out, as this sort of thing bothers me. Errors
* are never very well defined, and not easily dealt with.
*
* Changed to use InterlockedExchange to store function pointer.
* I doubt this is needed, but I've seen it used in this way (actually,
* I saw a critical section used). I tried simple assignment #ifndef _MT
* but the inline C function was more code (InterlockedExchange is about 5 instructions).
*
* Changed to use a reusable function to do the unicode \\?\ mucking.
*/
/*
 * a minimalist/no compromise clean/dirty implementation
 * of opendir/readdir/closedir for Windows 95 and NT.
 *
 * provides minimum information, but supports arbitrarily long paths.
 * follows *some* strict coding guidelines, but pulls many low C punches.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <windows.h>
#include "a2wetc.h"
#include <winerror.h>
#include <errno.h>
#include <malloc.h>
#include <wchar.h>
#include <tchar.h>
#define or ||
#define and &&
#include "gstring.h"
#include "cstring.h"
#include "dirent.h"
#include "w95.h"

#ifdef UNICODE
#define FindFirstFile FindFirstFileW95
#define FindNextFile FindNextFileW95
#endif

#undef ASSERT
#define ASSERT(x) assert(x)
#undef VERIFY
#define VERIFY(x) verify(x)

/* asserts are removed from release, verify still evals the code */
inline void verify(BOOL x)
{
	assert(x);
}
inline void verify(void* x)
{
	assert(x);
}

#pragma hdrstop

extern "C" BOOL ISDIRSEP(unsigned ch)
{
	return (ch == '/' || ch == '\\');
}

extern "C" BOOL ISABSOLUTE(const char* cp)
{
	const static char alpha[] = "\
abcdefghijklmnopqrstuvwxyz\
ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	if (ISDIRSEP(cp[0])) return TRUE;
	if (strchr(alpha, cp[0]) && cp[1] == ':' && ISDIRSEP(cp[2])) return TRUE;
	return FALSE;
}

//BOOL g_bReaddirConvertSlashesToForwardOnOutput = TRUE;

struct DIR {
	TCHAR* d_name; /* tricky: overlay with struct dirent */
	WIN32_FIND_DATA find_data;
	HANDLE handle;
};

static DIR *opendirA(LPCSTR path);
static DIR *opendirW(LPCWSTR path);

static void translate_error();

#if 0
class Path
// on NT, prepends appropriate \\?\ or \\?\UNC prefix
// maintains wchar_t and char.
{
public:
	Path(const char*);
	Path(const wchar_t*);
	operator const char*();
	operator const wchar_t*();
};
#endif

static CString UnlimitedLengthUnicodePathWithWackyPrefixIfNecessary(LPCWSTR wpath);
static CString UnlimitedLengthUnicodePathWithWackyPrefixIfNecessary(LPCSTR path);

static CString UnlimitedLengthUnicodePathWithWackyPrefixIfNecessary(LPCSTR path)
{
	if (path == NULL or path[0] == 0)
	{
		return path;
	}
	USES_CONVERSION;
	return UnlimitedLengthUnicodePathWithWackyPrefixIfNecessary(A2W(path));
}

static CString UnlimitedLengthUnicodePathWithWackyPrefixIfNecessary(LPCWSTR wpath)
{
/* If the path given is already a fullpath, or short, it is returned unchanged.
* The inconsistency introduced is that the return value may or may not be
* a full path. CONSIDER making it always do the mucking, for an inefficiency
* in the common case -- short paths.
*/
	if (wpath == NULL or wpath[0] == 0)
	{
		return wpath;
	}
	int len = length(wpath);
	if (len < MAX_PATH)
	{
		return wpath;
	}
	wchar_t ch0 = wpath[0];
	wchar_t ch1 = wpath[1];
	if ((ch0 == '\\' || ch0 == '/') && (ch1 == '\\' || ch1 == '/'))
	{
		return wpath;
	}
	wchar_t ch2 = wpath[2];
	if (iswalpha(ch0) && (ch1 == ':') && (ch2 == '\\' || ch2 == '/'))
	{
		return wpath;
	}
	len = GetFullPathNameW(wpath, 0, NULL, NULL); /* find length+1 of absolute path */
// returned len includes trailing nul? It seems so.
	if (len == 0) /* error */
	{
		translate_error();
		ASSERT(!"bad path in opendirw");
		return _T("");
	}
	if (len < MAX_PATH)
	{
		// might've parsed out redundancies, like x/../x/../
		// might be off by one..
		return wpath;
	}
	// X:\ => "\\?\X:\"
	//         1234
	// \\computer\share\path => \\?\UNC\computer\share\path
	//                            123456
	//                          1234567
	wchar_t* fullbuf;
	fullbuf = (wchar_t*) alloca(sizeof(*fullbuf)*(len+6));
	ASSERT(fullbuf != NULL);
	if (fullbuf == NULL)
	{
		return _T("");
	}
	// get absolute path, leaving room for \\?\ or \\?\UNC
	int len2 = GetFullPathNameW(wpath, len, fullbuf+6, NULL);
// This len does not include trailing nul.
	if (len2 == 0)
	{ /* error */
		translate_error();
		ASSERT(!"GetFullPathNameW failed");
		return _T("");
	}
	ASSERT(len2+1 == len);
	len = len2;
	ASSERT(wpath[len] == 0);
	wchar_t firstchar = fullbuf[6+0];
	wchar_t secondchar = firstchar ? fullbuf[6+1] : 0;
	ASSERT(firstchar != 0);
	ASSERT(secondchar != 0); // no such thing as single char full path in windows
	// minimum directory full paths are A:\ and \\A
	// but the validity of \\A is unclear, it might have to be \\A\B or "\\A\B\"
	// (quotes prevent line continuation)
	ASSERT(fullbuf[6+2] != 0);
	unsigned char adjust=0;
	if ( // beware of ctype here?
		((firstchar >= 'A' && firstchar <= 'Z')
		|| (firstchar >= 'a' && firstchar <= 'z'))
		&& secondchar == ':')
	{
		ASSERT(fullbuf[6+2] == '\\');
		// not unc
		adjust = 2;
		memcpy(fullbuf+adjust, L"\\\\?\\", 4*sizeof(*fullbuf));
	}
	else if (firstchar == '\\' && secondchar =='\\') // allow '/' ?
	{ // (GetFullPathName tends to translate slashes.)
		// unc
		// "\\computer\share" becomes "\\?\UNC\computer\share"
		// notice the C overwrites a slash
		memcpy(fullbuf, L"\\\\?\\UNC", 7*sizeof(*fullbuf));
	}
	else
	{
		ASSERT(!"bad full path in opendirw");
		return _T("");
	}
	return fullbuf+adjust;
}

static void translate_error()
// attempt some errno translations
// similar to crt, subset.
// Just guesses. Not tested.
{
#ifdef _MSC_VER
// possibly Visual C++ specific
	DWORD win32er = GetLastError();
	int erno=-1;
	switch (win32er)
	{
		case ERROR_PATH_NOT_FOUND:
		case ERROR_FILE_NOT_FOUND:
		case ERROR_BAD_NETPATH:
		case ERROR_DEV_NOT_EXIST:
		case ERROR_INVALID_DRIVE:
		case ERROR_BAD_UNIT:
		case ERROR_BAD_NET_NAME:
			erno = ENOENT; break;

		case ERROR_NOT_ENOUGH_MEMORY:
		case ERROR_OUTOFMEMORY:
			erno = ENOMEM; break;

		case ERROR_ACCESS_DENIED:
			erno = EACCES; break;

		case ERROR_READ_FAULT:
			erno = EIO; break;
		case ERROR_NETWORK_BUSY:
			erno = EBUSY; break;
		default:;
	}
	errno = erno;
#endif
}

extern "C" DIR *opendir(LPCTSTR path)
{
	if (path == 0 || path[0] == 0)
	{
		return 0;
	}
	USES_CONVERSION;
	return opendirW(T2CW(path));
}

static DIR* opendirW(LPCWSTR path)
{
	CString fullpath = UnlimitedLengthUnicodePathWithWackyPrefixIfNecessary(path);
	if (fullpath.length() == 0)
	{
		return NULL;
	}
	if (fullpath[fullpath.length()-1] != '/' && fullpath[fullpath.length()-1] != '\\')
	{
		fullpath += TEXT("/");
	}
	fullpath += TEXT("*");
	DIR* dir = new DIR;
	if (dir == NULL)
	{
		return NULL;
	}
	USES_CONVERSION;
	HANDLE handle = FindFirstFile(fullpath, &dir->find_data);
	if (handle == INVALID_HANDLE_VALUE)
	{
		translate_error();
		delete dir;
		return NULL;
	}
	dir->handle = handle;
	dir->d_name = NULL; /* indicate next is first */
	return dir;
} /* opendirW */

extern "C" void closedir(DIR* dir)
{
	if (dir != NULL)
	{
		FindClose(dir->handle);
		delete dir;
	}
}

extern "C" struct dirent* readdir(DIR* dir)
{
	struct dirent* result;
	if (dir == NULL)
	{
		return NULL;
	}
	if (dir->d_name == NULL) /* return result of FindFirstFile */
	{
		dir->d_name = dir->find_data.cFileName;
		result = (struct dirent*) dir; /* DIR and struct dirent have same first field */
	}
	else
	{
		result = FindNextFile(dir->handle, &dir->find_data)
			? (struct dirent*) dir
			: NULL;
	}
	//if (result && g_bReaddirConvertSlashesToForwardOnOutput)
	//{
// We are only returning file names, not full paths.
		//TCHAR* c;
// CONSIDER use _tcschr
		//for (c = dir->d_name ; *c ; c++ )
		//{
		//	if (*c == '\\')
		//		*c = '/';
		//}
	///* translate \\ to / */
	//	LPTSTR back=result;
	//	while (back = _tcschr(back, '\\'))
	//	{
	//		*back = '/';
	//	}
	//}
	return result;
}

#ifdef TEST
#include <stdio.h>
int main(int argc, char** argv)
{
	DIR* d = opendir(argv[1]);
	if (d != NULL)
	{
		struct dirent* de = readdir(d);
		while (de != NULL)
		{
			printf("%s\n", de->d_name);
			de = readdir(d);
		}
	}
	return 0;
}
#endif /* TEST */

#endif /* _WIN32 */
